subroutine bicgstab(vals,    & ! The nnz double precision nonzeros in the matrix A
                    rows,    & ! The rows indices,   length nnz
                    cols,    & ! The column indices, length nnz
                    nrow,    & ! Number of rows in the matrix
                    ncol,    & ! Number of columns in the matrix
                    nnz,     & ! Number of nonzeros in A
                    x,       & ! Solution vector
                    b,       & ! Right hand side vector
                    tol,     & ! Upper bound on relative residual
                    maxits,  & ! Maximum allowed number of iterations
                    relres,  & ! Relative residual for each iteration
                    timings, & ! Array of timings for parts of bicgstab
                    iflag)     ! Error flag
!--------------------------------------------------------------------------------
!
! Subroutine to run the biconjugate gradients stabilized (AKA bicgstab) algorithm
! on a sparse matrix in CSR format. The matrix is stored the data structure
! given by the variables vals, rows, cols, nrow, ncol, nnz. This subprogram uses the
! BLAS whenever practical. However, "double daxpy" updates, like
!
!   q = r + omega*(q - alpha*v)
!
! where q, r, and v are vectors, do not call daxpy because it would require
! traversing the vectors at least twice, while the computation can be just by
! traversing them once. Automatic arrays are used for most of the temporary
! arrays; see ca. line 92.
!
! All of the spelled out Greek letters and auxillary vectors come straight from
! the usual algorithm statement. In general, ones with names that end in "h" are
! for the transpose Krylov subspace. The original spec used hats to distinguish them,
! the reason for omegah and betah. As for alpha, that is a built-in function in
! Fortran, so I spelled it "alfa" instead to avoid potential name collisions.
!
! Although it is unlikely for small scale computing, a strong possibility exists
! for indices to exceed 2 billion = 2^31 sometime soon. For that reason
! kindlytypes defines a long integer which is what mosts of the integers are
! defined as. Integers used in intrinsic function calls are mostly left to the
! default integer size, currently (2012) 32-bits on most platforms.
!
! The times for initialization, iteration, and mat-vec multiply are returned
! in the array timings. The relative residual norm ||r||/||b|| is stored in
! the array relres for each iteration.
!
! Since each iteration takes 2 mat-vec multiplications. the number of iterations
! returned in iflag is multiplied by 2 to give a more fair comparison with other 
! iterative methods.
!
!--------------
! On return: 
!--------------
!
! The error flag has following values: 
!
!    iflag =  0, allegedly successful completion
!    iflag = -k, the k-th input arguments was fouled up (k starts from 1, not 9)
!    iflag =  k, failure was encountered on iteration k
!    iflag = 2*maxits, maximum iterations were reached before reaching the
!            specified tolerance
!
! x holds the latest estimate of the solution to Ax = b.
! maxits is reset to be the number of iterations actually made
! b, tol, vals, rows, cols are not changed by this subroutine
!
!--------------------------------------------
! Randall Bramley
! Department of Computer Science
! Indiana University, Bloomington
!--------------------------------------------------------------------------------

    ! Defines integer and real kinds 
    use kindlytypes, only: int4 => integer4, int8 => integer8, real8
    use iso_fortran_env, only: stdout => output_unit, &
                               stderr => error_unit
    use noisy
    implicit none

    integer(kind=int4), intent(in) :: nrow, ncol, nnz
    integer(kind=int4) :: iflag, maxits, k
    integer(kind=int4) :: n

    ! Automatic arrays
    real(kind=real8), dimension(1:nrow) :: r, q, v, p, s, t

    ! Other stuff
    integer(kind=int4), intent(in) :: rows(nnz), cols(nnz)
    real(kind=real8), dimension(1:maxits) :: relres
    real(kind=real8), intent(in)   :: b(nrow), vals(nnz), tol
    real(kind=real8), intent(inout):: x(ncol)
    real(kind=real8) :: alfa, beta, betah, omega, omegah, sigma, tau
    real(kind=real8) :: rho, SMALL
    real(kind=real8) :: bnorm, time_matvec
    real(kind=real8), external :: ddot, elapsedtime
    real(kind=real8), dimension(:) :: timings

    !===================================
    ! Check for clearly dumb settings 
    !===================================
    iflag = 0
    ! Error flag values are the negative of the argument number that seems wrong 
    if (maxits < 1) iflag   = -10;
    if (nrow /= ncol) iflag = -4
    if (tol <= 0) iflag     = -9
    if (nnz < nrow) iflag   = -6
    if (iflag /=0) return

    SMALL = 1.0_real8/sqrt(huge(1.0_real8))
    n = nrow  ! nrow = ncol = n for a single-processor solver

    !==================
    ! Initialization
    !==================
    if (benoisy > 2)  write(stdout,600) 'In bicgstab, b = ', b
    bnorm = sqrt(ddot(n, b, 1, b, 1))
    if (benoisy > 1)  write(stdout,400) 'In bicgstab, bnorm = ', bnorm
    if (bnorm <= SMALL) then
        ! In this case could just set x = 0 and return
        bnorm = 1.0
        if (benoisy > 1) write(stdout,400) '       bnorm reset = ', bnorm
    end if

    ! Start the clocks
    timings(1) = elapsedtime()
    timings(3) = 0.0

    k = 0  ! k is the iteration counter
    if (benoisy > 2) write(stdout,600) 'In bicgstab, x = ', x

    ! Compute the residual vector using v = b - A*x
    time_matvec = elapsedtime()
    call CSRmatvec(vals, rows, cols, nrow, ncol, nnz, x, v)
    time_matvec = elapsedtime() - time_matvec
    timings(3) = timings(3) + time_matvec 
    v(1:n)  = b(1:n) - v(1:n)
    if (benoisy > 2) write(stdout,600) 'In bicgstab, v = ', v 

    ! Initial residual norm squared
    rho     = ddot(n, v, 1, v, 1) ! rho is now ||residual||^2
    relres(k+1) = sqrt(rho)/bnorm
    if (benoisy > 1) then
        write(stdout,400) '  initial rho = ', rho
        write(stdout,400) '  initial bnorm = ', bnorm
        write(stdout,400) '  initial relres(k+1) = ', relres(k+1)
    end if

    ! Set everything else
    r(1:n) = v(1:n)
    p(1:n) = r(1:n)
    v(1:n) = 0.0
    q(1:n) = 0.0
    omegah = 1.0
    beta   = 1.0
    alfa   = 1.0
    timings(1) = elapsedtime() - timings(1) 

    ! Iteration  
    timings(2) = elapsedtime() 
    iterate: do while (k < maxits .and. sqrt(rho)/bnorm > tol)

        k = k + 1
        betah = ddot(n, p, 1, r, 1)
        if (abs(beta) <= SMALL) then
            iflag = -2*k
            exit  ! exit from the while-loop, not just this if-block
        end if
    
        omega  = (betah/beta)*(omegah/alfa)
        beta   = betah
        q(1:n) = r(1:n) + omega*(q(1:n)-alfa*v(1:n))

        time_matvec = elapsedtime()
        call CSRmatvec(vals, rows, cols, nrow, ncol, nnz, q, t)
        time_matvec = elapsedtime() - time_matvec
        timings(3) = timings(3) + time_matvec 

        v(1:n) = t(1:n)
        tau    = ddot(n, p, 1, v, 1)
        if (benoisy > 2) then
            write(stdout,600) 'In bicgstab, p = ', p
            write(stdout,600) 'In bicgstab, v = ', v
        end if

        if (benoisy > 1) write(stdout,400) 'In bicgstab, tau = ', tau

        if (abs(tau) <= SMALL) then
            iflag = 2*k -1
            exit ! algorithm breakdown; exit the loop
        end if 
        omegah = betah/tau
        s(1:n) = r(1:n) - omegah*v(1:n)
    
        time_matvec = elapsedtime()
        call CSRmatvec(vals, rows, cols, nrow, ncol, nnz, s, t)
        time_matvec = elapsedtime() - time_matvec
        timings(3) = timings(3) + time_matvec 
    
        sigma = ddot(n, t, 1, s, 1)
        tau   = ddot(n, t, 1, t, 1)
    
        ! Check for early termination. If so, update x and bail out.  
        if (abs(tau) <=  SMALL) then
            iflag = 2*k
            call daxpy(n, omegah, q, 1, x, 1)
            exit ! from the while-loop
        end if 
    
        ! Algorithm breaks down if alfa is too small.  
        alfa = sigma/tau
        if (abs(alfa) <=   SMALL) then
            iflag = -2*k
            exit  ! from the while-loop
        end if

        ! Update x and residual vector 
        x(1:n) = x(1:n) + omegah*q(1:n) + alfa*s(1:n)
        r(1:n) = s(1:n) - alfa*t(1:n)
        rho = ddot(n, r, 1, r, 1)

        ! Store the convergence history using the array relres to contain the
        ! relative residual norm on each iteration. This avoids having to do file I/O
        ! on every iteration.
        relres(k+1) = sqrt(rho)/bnorm
        if (benoisy > 1) then
            write(stdout,400) 'In bicgstab, rho = ', rho 
            write(stdout,400) 'In bicgstab, bnorm = ', bnorm 
            write(stdout,700) 'In bicgstab, k, relres(k+1) = ', k, relres(k+1) 
        end if 

    end do iterate
    timings(2) = elapsedtime() - timings(2)
    if (benoisy > 0) then
        write(stdout,400) 'In bicgstab, initialization time = ', timings(1)
        write(stdout,400) 'In bicgstab, iteration time = ', timings(2)
        write(stdout,400) 'In bicgstab, mat-vec time = ', timings(3)
    end if 
   
    maxits = k    ! Return actual number of iterations
    return

    100 format(a)
    200 format(a,i0)
    300 format(a,i0,a)
    400 format(a,es27.17)
    500 format(a,*(/,3x, i0))  ! infinite repeat descriptor
    600 format(a,*(/,3x, es27.17))
    700 format(a,i0, 3x, es27.17)


end subroutine bicgstab
